#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.property module

This module defines files properties which can be used to automatically handle all the magic
behind external files management; this includes blobs management and their references counting.
"""

from pyramid.events import subscriber
from pyramid.threadlocal import get_current_registry
from zope.interface import alsoProvides
from zope.lifecycleevent import IObjectRemovedEvent, ObjectAddedEvent, ObjectCreatedEvent, \
    ObjectRemovedEvent
from zope.location import locate
from zope.schema.interfaces import IField

from pyams_file.file import FileFactory
from pyams_file.interfaces import FILE_CONTAINER_ATTRIBUTES, IFile, IFileFieldContainer, IFileInfo
from pyams_utils.adapter import get_annotation_adapter
from pyams_utils.interfaces.form import NOT_CHANGED, TO_BE_DELETED


__docformat__ = 'restructuredtext'


_MARKER = object()


def get_instance_attributes(instance):
    """Get file attributes of given instance"""
    return get_annotation_adapter(instance, FILE_CONTAINER_ATTRIBUTES, set,
                                  notify=False, locate=False)


class FileProperty:
    """Property class used to handle files"""

    def __init__(self, field, name=None, klass=None, **args):
        if not IField.providedBy(field):
            raise ValueError("Provided field must implement IField interface...")
        if name is None:
            name = field.__name__
        self.__field = field
        self.__name = name
        self.__klass = klass
        self.__args = args

    def __get__(self, instance, klass):
        if instance is None:
            return self
        value = instance.__dict__.get(self.__name, _MARKER)
        if value is _MARKER:
            field = self.__field.bind(instance)
            value = getattr(field, 'default', _MARKER)
            if value is _MARKER:
                raise AttributeError(self.__name)
        return value

    def __set__(self, instance, value):  # pylint: disable=too-many-branches
        if value is NOT_CHANGED:
            return
        registry = get_current_registry()
        if (value is not None) and (value is not TO_BE_DELETED):
            filename = None
            # file upload data converter returns a tuple containing
            # filename and buffered IO stream extracted from FieldStorage...
            if isinstance(value, tuple):
                filename, value = value
            # initialize file through factory
            if not IFile.providedBy(value):
                factory = self.__klass or FileFactory
                file = factory(value, **self.__args)
                registry.notify(ObjectCreatedEvent(file))
                if not file.get_size():
                    value.seek(0)  # because factory may read until end of file...
                    file.data = value
                value = file
            if filename is not None:
                info = IFileInfo(value)
                if info is not None:
                    info.filename = filename
        field = self.__field.bind(instance)
        field.validate(value)
        if field.readonly and instance.__dict__.has_key(self.__name):
            raise ValueError(self.__name, "Field is readonly")
        old_value = instance.__dict__.get(self.__name, _MARKER)
        if old_value != value:
            # check for previous value
            if (old_value is not _MARKER) and (old_value is not None):
                registry.notify(ObjectRemovedEvent(old_value))
            if value is TO_BE_DELETED:
                if self.__name in instance.__dict__:
                    del instance.__dict__[self.__name]
                attributes = get_instance_attributes(instance)
                if attributes and (self.__name in attributes):
                    attributes.remove(self.__name)
            else:
                # set name of new value
                name = '++attr++{0}'.format(self.__name)
                if value is not None:
                    locate(value, instance, name)
                instance.__dict__[self.__name] = value
                # store file attributes of instance
                if not IFileFieldContainer.providedBy(instance):
                    alsoProvides(instance, IFileFieldContainer)
                attributes = get_instance_attributes(instance)
                attributes.add(self.__name)
                registry.notify(ObjectAddedEvent(value, instance, name))

    def __delete__(self, instance):
        old_value = instance.__dict__.get(self.__name, _MARKER)
        if (old_value is not _MARKER) and (old_value is not None):
            registry = get_current_registry()
            registry.notify(ObjectRemovedEvent(old_value))
        attributes = get_instance_attributes(instance)
        if attributes and (self.__name in attributes):
            attributes.remove(self.__name)
        del instance.__dict__[self.__name]


class I18nFileProperty:
    """I18n property class used to handle files"""

    def __init__(self, field, name=None, klass=None, **args):
        if not IField.providedBy(field):
            raise ValueError("Provided field must implement IField interface...")
        if name is None:
            name = field.__name__
        self.__field = field
        self.__name = name
        self.__klass = klass
        self.__args = args

    def __get__(self, instance, klass):
        if instance is None:
            return self
        value = instance.__dict__.get(self.__name, _MARKER)
        if value is _MARKER:
            field = self.__field.bind(instance)
            value = getattr(field, 'default', _MARKER)
            if value is _MARKER:
                raise AttributeError(self.__name)
        return value

    def __set__(self, instance, value):
        # pylint: disable=too-many-locals,too-many-statements,too-many-branches
        registry = get_current_registry()
        for lang in value:
            lang_value = value[lang]
            if (lang_value is TO_BE_DELETED) or (lang_value is NOT_CHANGED):
                continue
            if lang_value is not None:
                filename = None
                # file upload data converter returns a tuple containing
                # filename and buffered IO stream extracted from FieldStorage...
                if isinstance(lang_value, tuple):
                    filename, lang_value = lang_value
                # initialize file through factory
                if not IFile.providedBy(lang_value):
                    factory = self.__klass or FileFactory
                    file = factory(lang_value, **self.__args)
                    registry.notify(ObjectCreatedEvent(file))
                    if not file.get_size():
                        lang_value.seek(0)  # because factory may read until end of file...
                        file.data = lang_value
                    lang_value = file
                if filename is not None:
                    info = IFileInfo(lang_value)
                    if info is not None:
                        info.filename = filename
            value[lang] = lang_value
        field = self.__field.bind(instance)
        field.validate(value)
        if field.readonly and instance.__dict__.has_key(self.__name):
            raise ValueError(self.__name, "Field is readonly")
        old_value = instance.__dict__.get(self.__name, _MARKER)
        if old_value != value:
            # check for previous value
            if old_value is _MARKER:
                old_value = {}
            for lang in value:
                new_lang_value = value.get(lang)
                if new_lang_value is NOT_CHANGED:
                    continue
                old_lang_value = old_value.get(lang, _MARKER)
                if (old_lang_value is not _MARKER) and (old_lang_value is not None):
                    registry.notify(ObjectRemovedEvent(old_lang_value))
                attrname = '{0}::{1}'.format(self.__name, lang)
                if new_lang_value is TO_BE_DELETED:
                    if (self.__name in instance.__dict__) and (lang in old_value):
                        del old_value[lang]
                    attributes = get_instance_attributes(instance)
                    if attributes and (attrname in attributes):
                        attributes.remove(attrname)
                else:
                    # set name of new value
                    name = '++i18n++{0}:{1}'.format(self.__name, lang)
                    if new_lang_value is not None:
                        locate(new_lang_value, instance, name)
                    old_value[lang] = new_lang_value
                    # store file attributes of instance
                    if not IFileFieldContainer.providedBy(instance):
                        alsoProvides(instance, IFileFieldContainer)
                    attributes = get_instance_attributes(instance)
                    attributes.add(attrname)
                    registry.notify(ObjectAddedEvent(new_lang_value, instance, name))
            instance.__dict__[self.__name] = old_value

    def __delete__(self, instance):
        attributes = get_instance_attributes(instance)
        old_value = instance.__dict__.get(self.__name, _MARKER)
        if (old_value is not _MARKER) and (old_value is not None):
            registry = get_current_registry()
            for lang in list(old_value):
                old_lang_value = old_value.get(lang, _MARKER)
                if (old_lang_value is not _MARKER) and (old_lang_value is not None):
                    registry.notify(ObjectRemovedEvent(old_lang_value))
                    del old_value[lang]
                attrname = '{0}::{1}'.format(self.__name, lang)
                if attributes and (attrname in attributes):
                    attributes.remove(attrname)
        del instance.__dict__[self.__name]


@subscriber(IObjectRemovedEvent, context_selector=IFileFieldContainer)
def handle_removed_file_container(event):
    """Handle removal of files containers"""
    instance = event.object
    attributes = get_instance_attributes(instance)
    for attr in attributes.copy():
        if '::' in attr:
            attr, lang = attr.split('::')  # pylint:disable=unused-variable
        delattr(instance, attr)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.skin.view module

This module provides a Pyramid view used to download any file.
"""

from http.client import NOT_MODIFIED, PARTIAL_CONTENT

from pyramid.response import Response
from pyramid.view import view_config
from zope.dublincore.interfaces import IZopeDublinCore

from pyams_file.interfaces import IFile
from pyams_utils.unicode import translate_string


__docformat__ = 'restructuredtext'


MAX_RANGE_LENGTH = 1 << 21  # 2 Mb


@view_config(context=IFile)
def FileView(request):  # pylint: disable=invalid-name
    """Default file view"""
    context = request.context

    # set content type
    content_type = context.content_type
    if isinstance(content_type, bytes):
        content_type = content_type.decode('utf-8')

    # check for last modification date
    response = Response(content_type=content_type)
    zdc = IZopeDublinCore(context, None)
    if zdc is not None:
        modified = zdc.modified
        if modified is not None:
            if_modified_since = request.if_modified_since
            # pylint: disable=no-member
            if if_modified_since and \
                    (int(modified.timestamp()) <= int(if_modified_since.timestamp())):
                return Response(content_type=content_type,
                                status=NOT_MODIFIED)
            response.last_modified = modified

    body_file = context.get_blob(mode='c')

    if request.params.get('download') is not None:
        filename = context.filename or 'noname.txt'
        response.content_disposition = 'attachment; filename="{0}"'.format(
            translate_string(filename, force_lower=False))

    # check for range request
    if request.range is not None:
        try:
            body = body_file.read()
            body_length = len(body)
            range_start = request.range.start or 0
            if 'Firefox' in request.user_agent:  # avoid partial range for Firefox videos
                range_end = body_length
            else:
                range_end = request.range.end or min(body_length, range_start + MAX_RANGE_LENGTH)
            ranged_body = body[range_start:range_end]
            response.status = PARTIAL_CONTENT
            response.headers['Content-Range'] = 'bytes {first}-{last}/{len}'.format(
                first=range_start, last=range_start + len(ranged_body) - 1, len=body_length)
            response.body = ranged_body
        finally:
            body_file.close()
    else:
        response.body_file = body_file

    return response



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.interfaces.thumbnail module

This module provides all thumbnails related interfaces.
"""

from zope.interface import Attribute, Interface
from zope.schema import Int


__docformat__ = 'restructuredtext'


class IThumbnailFile(Interface):
    """Marker interface for thumbnails"""


class IThumbnailGeometry(Interface):
    """Image thumbnail geometry interface"""

    # pylint: disable=invalid-name
    x1 = Int(title="Thumbnail position X1",
             required=True,
             min=0)

    # pylint: disable=invalid-name
    y1 = Int(title="Thumbnail position Y1",
             required=True,
             min=0)

    # pylint: disable=invalid-name
    x2 = Int(title="Thumbnail position X2",
             required=True,
             min=0)

    # pylint: disable=invalid-name
    y2 = Int(title="Thumbnail position Y2",
             required=True,
             min=0)

    def is_empty(self):
        """Check if geometry is not empty"""


class IThumbnailer(Interface):
    """Interface of adapter used to generate image thumbnails"""

    label = Attribute("Thumbnail label")
    section = Attribute("Thumbnail section")
    weight = Attribute("Thumbnail weight")

    def get_default_geometry(self):
        """Get default thumbnail geometry"""

    # pylint:disable=redefined-builtin
    def create_thumbnail(self, target, format=None):
        """Create thumbnail of the given source object

        Source can be any file which can provide thumbnails (image, video,
        PDF file...).
        Target, which defines thumbnail size, can be defined as a selection name
        ('pano', 'square', 'xs'...), as a geometry or as a (width, height) tuple.

        If the requested image is of a resolution higher than that of the original file,
        the resulting image resolution will be that of the original file.

        If format (JPEG, PNG...) is given, this will be the format of the generated
        thumbnail; otherwise the format will be those of the source image.
        """


class IThumbnails(Interface):
    """Image thumbnail interface

    Displays are images thumbnails generated 'on the fly' and stored into image
    annotations for future use
    """

    def get_image_size(self):
        """Get original image size"""

    def get_thumbnail_size(self, thumbnail_name, forced=False):
        """Get real size of the genrated thumbnail

        If forced is True, the generated thumbnail can be larger than the original
        source
        """

    def get_geometry(self, selection_name):
        """Get geometry of a given thumbnail"""

    def set_geometry(self, selection_name, geometry):
        """Set geometry for given thumbnail"""

    def clear_geometries(self):
        """Remove all stored geometries from object annotations"""

    def get_thumbnail_name(self, thumbnail_name, with_size=None):
        """Get matching name for the given thumbnail name or size"""

    # pylint:disable=redefined-builtin
    def get_selection(self, selection_name, format=None):
        """Get image for given user selection"""

    # pylint:disable=redefined-builtin,too-many-arguments
    def get_thumbnail(self, thumbnail_name, format=None, watermark=None,
                      watermark_position='scale', watermark_opacity=1):
        """Get requested thumbnail

        Display can be specified as:
        - a name matching a custom thumbnailer utility
        - a width, as wXXX where XXX is the requested image width
        - a height, as hYYY, where YYY is the requested image height
        - a size, as XXXxYYY
        """

    def delete_thumbnail(self, thumbnail_name):
        """Remove selected thumbnail from object annotations"""

    def clear_thumbnails(self):
        """Remove all thumbnails from object annotations"""


class IWatermarker(Interface):
    """Interface of utility used to add image watermark"""

    # pylint:disable=redefined-builtin,too-many-arguments
    def add_watermark(self, image, watermark, position='scale', opacity=1, format=None):
        """Add watermark to given image"""



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.archive module

Base archive extractor class
"""


from io import BytesIO


class ArchiveExtractorBase:
    """Base archive extractor utility class"""

    def _initialize(self, data, mode='r'):  # pylint: disable=no-self-use,unused-argument
        """Initialize extractor data"""
        if isinstance(data, tuple):
            data = data[0]
        if isinstance(data, str):
            data = data.encode()
        if isinstance(data, bytes):
            data = BytesIO(data)
        return data



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.image module

This module provides several adapter to handle images thumbnails generation.

"""

import re
from io import BytesIO

from PIL import Image, ImageFilter
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty

from pyams_file.interfaces import IResponsiveImage, IImageFile
from pyams_file.interfaces.thumbnail import IThumbnailGeometry, IThumbnailer, IThumbnails
from pyams_utils.adapter import ContextAdapter, adapter_config


__docformat__ = 'restructuredtext'

from pyams_file import _  # pylint: disable=ungrouped-imports


WEB_FORMATS = ('JPEG', 'PNG', 'GIF')
THUMB_SIZE = re.compile(r'^(?:\w+:)?([0-9]+)x([0-9]+)$')


@implementer(IThumbnailGeometry)
class ThumbnailGeometry:
    """Image thumbnail geometry"""

    x1 = FieldProperty(IThumbnailGeometry['x1'])  # pylint: disable=invalid-name
    y1 = FieldProperty(IThumbnailGeometry['y1'])  # pylint: disable=invalid-name
    x2 = FieldProperty(IThumbnailGeometry['x2'])  # pylint: disable=invalid-name
    y2 = FieldProperty(IThumbnailGeometry['y2'])  # pylint: disable=invalid-name

    def __repr__(self):
        return '<ThumbnailGeometry: x1,y1={0.x1},{0.y1} - x2,y2={0.x2},{0.y2}>'.format(self)

    def __eq__(self, other):
        if IThumbnailGeometry.providedBy(other):
            return (self.x1 == other.x1) and \
                   (self.x2 == other.x2) and \
                   (self.y1 == other.y1) and \
                   (self.y2 == other.y2)
        return False

    def is_empty(self):
        """Check if geometry is empty"""
        return (self.x2 <= self.x1) or (self.y2 <= self.y1)


@adapter_config(context=IImageFile, provides=IThumbnailer)
class ImageThumbnailer(ContextAdapter):
    """Image thumbnailer adapter"""

    label = _("Default thumbnail")
    section = _("Default thumbnail")
    weight = 1

    def get_default_geometry(self, options=None):    # pylint: disable=unused-argument
        """Default thumbnail geometry"""
        geometry = ThumbnailGeometry()
        width, height = self.context.get_image_size()
        geometry.x1 = 0  # pylint: disable=invalid-name
        geometry.y1 = 0  # pylint: disable=invalid-name
        geometry.x2 = width  # pylint: disable=invalid-name
        geometry.y2 = height  # pylint: disable=invalid-name
        return geometry

    def create_thumbnail(self, target, format=None):
        # pylint: disable=redefined-builtin,too-many-branches
        """Create thumbnail of given size and format

        :param target: size or geometry of the thumbnail
        :param format: image format
        """
        # check thumbnail name
        if isinstance(target, str):
            width, height = tuple(map(int, target.split('x')))
        elif IThumbnailGeometry.providedBy(target):
            width = target.x2 - target.x1
            height = target.y2 - target.y1
        elif isinstance(target, tuple):
            width, height = target
        else:
            return None
        # check format
        with self.context as blob:
            try:
                if blob is None:
                    return None
                image = Image.open(blob)
                if not format:
                    format = image.format
                format = format.upper()
                if format not in WEB_FORMATS:
                    format = 'JPEG'
                # check image mode
                if image.mode in ('P', 'RGBA'):
                    if format == 'JPEG':
                        image = image.convert('RGB')
                    elif image.mode != 'RGBA':
                        image = image.convert('RGBA')
                # generate thumbnail
                new_image = BytesIO()
                image.resize((width, height), Image.ANTIALIAS) \
                    .filter(ImageFilter.UnsharpMask(radius=0.5, percent=100, threshold=0)) \
                    .save(new_image, format)
                return new_image, format.lower()
            finally:
                if blob is not None:
                    blob.close()


class ImageSelectionThumbnailer(ImageThumbnailer):
    """Image thumbnailer based on user selection"""

    section = _("Custom selections")

    def create_thumbnail(self, target, format=None):
        # pylint: disable=redefined-builtin,too-many-branches
        # get thumbnail size
        if isinstance(target, str):
            # pylint: disable=assignment-from-no-return
            geometry = IThumbnails(self.context).get_geometry(target)
            match = THUMB_SIZE.match(target)
            if match:
                width, height = tuple(map(int, match.groups()))
            else:
                width = abs(geometry.x2 - geometry.x1)
                height = abs(geometry.y2 - geometry.y1)
        elif IThumbnailGeometry.providedBy(target):
            geometry = target
            width = abs(geometry.x2 - geometry.x1)
            height = abs(geometry.y2 - geometry.y1)
        elif isinstance(target, tuple):
            width, height = target
            geometry = self.get_default_geometry()
        else:
            return None
        # check format
        with self.context as blob:
            try:
                if blob is None:
                    return None
                image = Image.open(blob)
                if not format:
                    format = image.format
                format = format.upper()
                if format not in WEB_FORMATS:
                    format = 'JPEG'
                # check image mode
                if image.mode in ('P', 'RGBA'):
                    if format == 'JPEG':
                        image = image.convert('RGB')
                    elif image.mode != 'RGBA':
                        image = image.convert('RGBA')
                # generate thumbnail
                new_image = BytesIO()
                thumb_size = self.get_thumb_size(width, height, geometry)
                image.crop((geometry.x1, geometry.y1, geometry.x2, geometry.y2)) \
                    .resize(thumb_size, Image.ANTIALIAS) \
                    .filter(ImageFilter.UnsharpMask(radius=0.5, percent=100, threshold=0)) \
                    .save(new_image, format)
                return new_image, format.lower()
            finally:
                if blob is not None:
                    blob.close()

    def get_thumb_size(self, width, height, geometry):
        # pylint: disable=no-self-use,unused-argument
        """Get thumbnail size based on given dimensions"""
        return width, height


class ImageRatioThumbnailer(ImageSelectionThumbnailer):
    """Image thumbnailer with specific ratio"""

    ratio = (None, None)  # (width, height) ratio tuple

    def get_default_geometry(self, options=None):
        """Get default geometry for selection"""
        geometry = ThumbnailGeometry()
        width, height = self.context.get_image_size()
        thumb_max_height = width * self.ratio[1] / self.ratio[0]
        if thumb_max_height >= height:
            # image wider
            thumb_width = height * self.ratio[0] / self.ratio[1]
            geometry.x1 = round((width / 2) - (thumb_width / 2))
            geometry.y1 = 0
            geometry.x2 = round((width / 2) + (thumb_width / 2))
            geometry.y2 = height
        else:
            thumb_height = thumb_max_height
            geometry.x1 = 0
            geometry.y1 = round((height / 2) - (thumb_height / 2))
            geometry.x2 = width
            geometry.y2 = round((height / 2) + (thumb_height / 2))
        return geometry


@adapter_config(name='portrait', context=IImageFile, provides=IThumbnailer)
class ImagePortraitThumbnailer(ImageRatioThumbnailer):
    """Image portrait thumbnail adapter"""

    label = _("Portrait thumbnail")
    weight = 5

    ratio = (3, 4)


@adapter_config(name='square', context=IImageFile, provides=IThumbnailer)
class ImageSquareThumbnailer(ImageRatioThumbnailer):
    """Image square thumbnail adapter"""

    label = _("Square thumbnail")
    weight = 6

    ratio = (1, 1)


class ImageHorizontalThumbnailer(ImageRatioThumbnailer):
    """Image horizontal thumbnailer"""

    def get_thumb_size(self, width, height, geometry):
        thumb_size = abs(geometry.x2 - geometry.x1), abs(geometry.y2 - geometry.y1)
        w_ratio = 1. * width / thumb_size[0]
        h_ratio = 1. * height / thumb_size[1]
        ratio = min(w_ratio, h_ratio)
        return round(ratio * thumb_size[0]), round(ratio * thumb_size[1])


@adapter_config(name='pano', context=IImageFile, provides=IThumbnailer)
class ImagePanoThumbnailer(ImageHorizontalThumbnailer):
    """Image panoramic thumbnail adapter"""

    label = _("Panoramic thumbnail")
    weight = 7

    ratio = (16, 9)


@adapter_config(name='card', context=IImageFile, provides=IThumbnailer)
class ImageCardThumbnailer(ImageHorizontalThumbnailer):
    """Image card thumbnail adapter"""

    label = _("Card thumbnail")
    weight = 8

    ratio = (2, 1)


@adapter_config(name='banner', context=IImageFile, provides=IThumbnailer)
class ImageBannerThumbnailer(ImageRatioThumbnailer):
    """Image banner thumbnail adapter"""

    label = _("Banner thumbnail")
    weight = 9

    ratio = (5, 1)


class ResponsiveImageThumbnailer(ImageSelectionThumbnailer):
    """Responsive image thumbnailer"""

    section = _("Responsive selections")


@adapter_config(name='xs', context=IResponsiveImage, provides=IThumbnailer)
class XsImageThumbnailer(ResponsiveImageThumbnailer):
    """eXtra-Small responsive image thumbnailer"""

    label = _("Smartphone thumbnail")
    weight = 10


@adapter_config(name='sm', context=IResponsiveImage, provides=IThumbnailer)
class SmImageThumbnailer(ResponsiveImageThumbnailer):
    """SMall responsive image thumbnailer"""

    label = _("Tablet thumbnail")
    weight = 11


@adapter_config(name='md', context=IResponsiveImage, provides=IThumbnailer)
class MdImageThumbnailer(ResponsiveImageThumbnailer):
    """MeDium responsive image thumbnailer"""

    label = _("Medium screen thumbnail")
    weight = 12


@adapter_config(name='lg', context=IResponsiveImage, provides=IThumbnailer)
class LgImageThumbnailer(ResponsiveImageThumbnailer):
    """LarGe responsive image thumbnailer"""

    label = _("Large screen thumbnail")
    weight = 13


@adapter_config(name='xl', context=IResponsiveImage, provides=IThumbnailer)
class XlImageThumbnailer(ResponsiveImageThumbnailer):
    """EXtra-Large responsive image thumbnailer"""

    label = _("Extra-large screen thumbnail")
    weight = 14



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.file module

This module is the core of PyAMS_file package. It defines a base :py:class:`File` class which is
used to store all external files as blobs.

The way in which files are physically stored can depend on ZODB storage settings, in particular
about the "shared blobs cache" parameter.
"""

try:
    import magic
except ImportError:
    magic = None

import os
from io import BytesIO

from BTrees.OOBTree import OOBTree  # pylint: disable=import-error,no-name-in-module
from PIL import Image
from ZODB.blob import Blob
from ZODB.utils import oid_repr
from persistent import Persistent
from pyramid.events import subscriber
from zope.container.contained import Contained
from zope.copy.interfaces import ICopyHook, ResumeCopy
from zope.interface import implementer
from zope.lifecycleevent import IObjectAddedEvent, IObjectRemovedEvent
from zope.location.interfaces import IContained
from zope.schema.fieldproperty import FieldProperty

from pyams_file.interfaces import FileModifiedEvent, IAudioFile, IBlobReferenceManager, IFile, \
    IFileInfo, IImageFile, ISVGImageFile, IVideoFile
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.registry import get_utility
from pyams_utils.request import check_request


__docformat__ = 'restructuredtext'


BLOCK_SIZE = 1 << 16

EXTENSIONS_THUMBNAILS = {
    '.7z': 'application-x-7z-compressed.png',
    '.ac3': 'audio-ac3.png',
    '.afm': 'application-x-font-afm.png',
    '.avi': 'video-x-generic.png',
    '.bmp': 'application-x-egon.png',
    '.bz2': 'application-x-bzip.png',
    '.css': 'text-css.png',
    '.csv': 'text-csv.png',
    '.doc': 'application-msword.png',
    '.docx': 'application-msword.png',
    '.dot': 'application-msword-template.png',
    '.deb': 'application-x-deb.png',
    '.eps': 'image-x-eps.png',
    '.exe': 'application-x-ms-dos-executable.png',
    '.flv': 'application-x-shockwave-flash.png',
    '.gif': 'application-x-egon.png',
    '.gz': 'application-x-bzip.png',
    '.htm': 'application-x-mswinurl.png',
    '.html': 'application-x-mswinurl.png',
    '.jar': 'application-x-java-archive.png',
    '.java': 'text-x-java.png',
    '.jpeg': 'application-x-egon.png',
    '.jpg': 'application-x-egon.png',
    '.js': 'application-javascript.png',
    '.mp2': 'audio-ac3.png',
    '.mp3': 'audio-ac3.png',
    '.mp4': 'application-x-shockwave-flash.png',
    '.mpeg': 'audio-ac3.png',
    '.mpg': 'audio-ac3.png',
    '.mov': 'application-x-shockwave-flash.png',
    '.odf': 'odf.png',
    '.odp': 'application-vnd.oasis.opendocument.presentation.png',
    '.ods': 'application-vnd.oasis.opendocument.spreadsheet.png',
    '.odt': 'application-msword.png',
    '.ogg': 'audio-x-flac+ogg.png',
    '.otf': 'application-x-font-otf.png',
    '.otp': 'application-vnd.oasis.opendocument.presentation-template.png',
    '.ots': 'application-vnd.oasis.opendocument.spreadsheet-template.png',
    '.ott': 'application-msword-template.png',
    '.pdf': 'application-pdf.png',
    '.php': 'application-x-php.png',
    '.pl': 'application-x-perl.png',
    '.png': 'application-x-egon.png',
    '.ppt': 'application-vnd.ms-powerpoint.png',
    '.ps': 'application-postscript.png',
    '.psd': 'application-x-krita.png',
    '.py': 'text-x-python.png',
    '.rpm': 'application-x-rpm.png',
    '.rdf': 'text-rdf+xml.png',
    '.rtf': 'application-rtf.png',
    '.sql': 'text-x-sql.png',
    '.svg': 'application-x-kontour.png',
    '.tif': 'application-x-egon.png',
    '.tiff': 'application-x-egon.png',
    '.ttf': 'application-x-font-ttf.png',
    '.txt': 'text-plain.png',
    '.vhd': 'application-x-smb-workgroup.png',
    '.xls': 'application-vnd.ms-excel.png',
    '.xlsx': 'application-vnd.ms-excel.png',
    '.xml': 'application-xml.png',
    '.wav': 'audio-x-adpcm.png',
    '.webm': 'application-x-shockwave-flash.png',
    '.wmf': 'application-x-wmf.png',
    '.wmv': 'video-x-generic.png',
    '.xcf': 'application-x-krita.png',
    '.zip': 'application-x-7z-compressed.png'
}


#
# Blobs references manager utility
#

@implementer(IBlobReferenceManager)
class BlobReferencesManager(Persistent, Contained):
    """Global blobs references manager utility

    The utility is used to keep all references of persistent files objects to
    their blobs; typically, when duplicating contents, we just increase blobs references instead
    of duplicating all their internal files contents, until they are modified.

    References management is done automatically when using file-related properties, like
    :py:class:`FileProperty <pyams_file.property.FileProperty>` or :py:class:`I18nFileProperty
    <pyams_file.property.I18nFileProperty>`.
    """

    def __init__(self):
        self.refs = OOBTree()

    def add_reference(self, blob, reference):
        """Add reference to given blob"""
        oid = getattr(blob, '_p_oid')
        if not oid:
            getattr(reference, '_p_jar').add(blob)
            oid = getattr(blob, '_p_oid')
        oid = oid_repr(oid)
        refs = self.refs.get(oid) or set()
        refs.add(reference)
        self.refs[oid] = refs

    def drop_reference(self, blob, reference):
        """Remove reference from given blob"""
        oid = oid_repr(getattr(blob, '_p_oid'))
        refs = self.refs.get(oid)
        if refs is not None:
            if reference in refs:
                refs.remove(reference)
            if refs:
                self.refs[oid] = refs
            else:
                del self.refs[oid]
                del blob
        else:
            del blob


#
# Persistent file class
#

@implementer(IFile, IFileInfo, IContained)
class File(Persistent, Contained):
    """Generic file persistent object"""

    title = FieldProperty(IFileInfo['title'])
    description = FieldProperty(IFileInfo['description'])
    filename = FieldProperty(IFileInfo['filename'])
    language = FieldProperty(IFileInfo['language'])

    _size = 0

    def __init__(self, data='', content_type=None, source=None):
        self.content_type = content_type
        self._blob = None
        if data:
            self.data = data
        elif source:
            if os.path.exists(source):
                try:
                    f = open(source, 'rb')  # pylint: disable=invalid-name
                    self.data = f
                finally:
                    f.close()

    def init_blob(self):
        """Initialize internal blob and add reference to it"""
        self.remove_blob_reference()
        self._blob = Blob()

    def add_blob_reference(self, reference=None):
        """Add reference to internal blob"""
        if self._blob is not None:
            references = get_utility(IBlobReferenceManager)
            references.add_reference(self._blob, reference if reference is not None else self)

    def remove_blob_reference(self):
        """Remove reference to internal blob

        Blob is deleted if there is no more reference to it.
        """
        if self._blob is not None:
            references = get_utility(IBlobReferenceManager)
            references.drop_reference(self._blob, self)
            self._blob = None

    def get_blob(self, mode='r'):
        """Open current blob in given mode"""
        if self._blob is None:
            return None
        return self._blob.open(mode=mode)

    def get_detached_blob(self):
        """Get a detached blob, which can be used after transaction end"""
        if self._blob is None:
            return None
        return open(self._blob.committed(), 'rb')

    def _get_data(self):
        """Read the whole blob content"""
        f = self.get_blob()  # pylint: disable=invalid-name
        if f is None:
            return None
        try:
            data = f.read()
            return data
        finally:
            f.close()

    def _set_data(self, data):
        """Set blob content"""
        self.init_blob()
        if isinstance(data, str):
            data = data.encode('utf-8')
        elif hasattr(data, 'seek'):
            data.seek(0)
        f = self.get_blob('w')  # pylint: disable=invalid-name
        try:
            if hasattr(data, 'read'):
                self._size = 0
                _data = data.read(BLOCK_SIZE)
                size = len(_data)
                while size > 0:
                    f.write(_data)
                    self._size += size
                    _data = data.read(BLOCK_SIZE)
                    size = len(_data)
            else:
                f.write(data)
                self._size = len(data)
        finally:
            f.close()

    data = property(_get_data, _set_data)

    def get_size(self):
        """Return current blob size"""
        return self._size

    def __enter__(self):
        """Context manager entry point"""
        return self.get_blob(mode='r')

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit point

        We don't close the blob here.
        It's up to the context manager call to close it!
        """

    def __iter__(self):
        """Iterator over blob content to read by chunks"""
        if self._blob is None:
            return
        with self as f:  # pylint: disable=invalid-name
            while True:
                chunk = f.read(BLOCK_SIZE)
                if not chunk:
                    f.close()
                    return
                yield chunk

    def __bool__(self):
        """Non-zero test against blob size"""
        return self._size > 0


@subscriber(IObjectAddedEvent, context_selector=IFile)
def handle_added_file(event):
    """Add blob reference when file is added"""
    event.object.add_blob_reference()


@subscriber(IObjectRemovedEvent, context_selector=IFile)
def handle_removed_file(event):
    """Remove blob associated with file when removed"""
    event.object.remove_blob_reference()


@adapter_config(context=IFile, provides=ICopyHook)
class BlobFileCopyHook(ContextAdapter):
    """Blob file copy hook

    Inspired by z3c.blobfile package. When a blob file is copied, we just add a reference to
    it instead of copying it's contents, until it is modified.
    """

    def __call__(self, toplevel, register):
        register(self._copy_blob)
        raise ResumeCopy

    def _copy_blob(self, translate):
        # Just add a reference to blob when copying file
        target = translate(self.context)
        setattr(target, '_blob', getattr(self.context, '_blob'))
        target.add_blob_reference(target)


#
# Persistent images
#

@implementer(IImageFile)
class ImageFile(File):
    """Image file persistent object"""

    image_size = (-1, -1)

    def _set_data(self, data):
        if isinstance(data, str):
            data = BytesIO(data.encode('utf-8'))
        elif isinstance(data, bytes):
            data = BytesIO(data)
        File._set_data(self, data)
        if hasattr(data, 'seek'):
            data.seek(0)
        img = Image.open(data)
        self.image_size = img.size

    data = property(File._get_data, _set_data)

    def get_image_size(self):
        """Get image dimensions"""
        return self.image_size

    def resize(self, width, height, keep_ratio=True):
        """Resize image to given dimensions"""
        with self as blob:
            try:
                image = Image.open(blob)
                image_size = image.size
                if width >= image_size[0] and height >= image_size[1]:
                    return
                new_image = BytesIO()
                w_ratio = 1. * width / image_size[0]
                h_ratio = 1. * height / image_size[1]
                if keep_ratio:
                    ratio = min(w_ratio, h_ratio)
                    image.resize((round(ratio * image_size[0]), round(ratio * image_size[1])),
                                 Image.ANTIALIAS) \
                         .save(new_image, image.format, quality=99)
                else:
                    image.resize((round(w_ratio * image_size[0]), round(h_ratio * image_size[1])),
                                 Image.ANTIALIAS) \
                         .save(new_image, image.format, quality=99)
                self.data = new_image
                request = check_request()
                request.registry.notify(FileModifiedEvent(self))
            finally:
                blob.close()

    def crop(self, x1, y1, x2, y2):  # pylint: disable=invalid-name
        """Crop image to given coordinates"""
        with self as blob:
            try:
                image = Image.open(blob)
                new_image = BytesIO()
                image.crop((x1, y1, x2, y2)) \
                     .save(new_image, image.format, quelity=99)
                self.data = new_image
                request = check_request()
                request.registry.notify(FileModifiedEvent(self))
            finally:
                blob.close()

    def rotate(self, angle=-90):
        """Rotate image of given angle, in clockwise degrees"""
        with self as blob:
            try:
                image = Image.open(blob)
                new_image = BytesIO()
                image.rotate(angle, expand=True) \
                     .save(new_image, image.format, quality=99)
                self.data = new_image
                request = check_request()
                request.registry.notify(FileModifiedEvent(self))
            finally:
                blob.close()


@implementer(ISVGImageFile)
class SVGImageFile(File):
    """SVG image file persistent object"""


@implementer(IVideoFile)
class VideoFile(File):
    """Video file persistent object"""


@implementer(IAudioFile)
class AudioFile(File):
    """Audio file persistent object"""


#
# Generic files utilities
#

def get_magic_content_type(input):  # pylint: disable=redefined-builtin
    """Get content-type based on magic library as *bytes*

    As libmagic bindings are provided via several 'magic' packages, we try them in order
    """
    if magic is not None:
        if hasattr(input, 'seek'):
            input.seek(0)
        if hasattr(input, 'read'):
            input = input.read()
        if hasattr(magic, 'detect_from_content'):
            result = magic.detect_from_content(input)  # pylint: disable=no-member
            if result:
                return result.mime_type
        elif hasattr(magic, 'from_buffer'):
            return magic.from_buffer(input, mime=True)
    return None


def FileFactory(data):  # pylint: disable=invalid-name
    """File object factory

    Automatically create the right file type based on magic
    content-type recognition.
    """
    content_type = get_magic_content_type(data)
    if content_type.startswith('image/svg'):
        factory = SVGImageFile
    elif content_type.startswith('image/'):
        factory = ImageFile
    elif content_type.startswith('video/'):
        factory = VideoFile
    elif content_type.startswith('audio/'):
        factory = AudioFile
    else:
        factory = File
    return factory(data, content_type)



#
# Copyright (c) 2008-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.generations sub-package

"""

import sys
from importlib import import_module

from pyams_file.file import BlobReferencesManager
from pyams_file.interfaces import IBlobReferenceManager
from pyams_site.generations import check_required_utilities
from pyams_site.interfaces import ISiteGenerations
from pyams_utils.registry import utility_config


__docformat__ = 'restructuredtext'


REQUIRED_UTILITIES = (
    (IBlobReferenceManager, '', BlobReferencesManager, 'Blobs references manager'),
)


@utility_config(name='PyAMS file', provides=ISiteGenerations)
class FilesGenerationsChecker:
    """PyAMS file package generations checker"""

    order = 40
    generation = 3

    def evolve(self, site, current=None):
        """Check for required utilities, tables and tools"""
        check_required_utilities(site, REQUIRED_UTILITIES)
        if not current:
            current = 1
        for generation in range(current, self.generation):
            module_name = 'pyams_file.generations.evolve{}'.format(generation)
            module = sys.modules.get(module_name)
            if module is None:
                module = import_module(module_name)
            module.evolve(site)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.skin main module

This module provides functions used to render images.
"""

from datetime import datetime

from pyquery import PyQuery
from pyramid.renderers import render
from zope.dublincore.interfaces import IZopeDublinCore

from pyams_file.interfaces import ISVGImageFile
from pyams_file.interfaces.thumbnail import IThumbnails
from pyams_utils.request import check_request
from pyams_utils.url import absolute_url


__docformat__ = 'restructuredtext'


def render_svg(image, width=None, height=None, request=None, css_class='', img_class='', alt=''):
    """Render SVG file"""
    # pylint: disable=too-many-arguments
    options = {}
    if width or height:
        options['style'] = 'width: {0}{1}; height: {2}{3};'.format(
            width, 'px' if isinstance(width, int) else '',
            height, 'px' if isinstance(height, int) else '')
    else:
        options['style'] = ''
    options['css_class'] = css_class
    if alt or img_class:
        svg = PyQuery(image.data)
        if alt:
            group = PyQuery('<g></g>')
            group.append(PyQuery('<title />').text(alt))
            for child in svg.children():
                group.append(child)
            svg.empty().append(group)
        if img_class:
            svg.attr('class', img_class)
        options['svg'] = svg.outer_html()
    else:
        options['svg'] = image.data
    return render('templates/svg-render.pt', options, request)


def render_img(image, width=None, height=None, request=None, css_class='', img_class='',
               timestamp=False, alt=''):
    # pylint: disable=too-many-arguments,assignment-from-no-return
    """Render image thumbnail"""
    thumbnail = None
    thumbnails = IThumbnails(image, None)
    if thumbnails is not None:
        if width and height:
            thumbnail = thumbnails.get_thumbnail('{0}x{1}'.format(width, height))
        elif width and (width != 'auto'):
            thumbnail = thumbnails.get_thumbnail('w{0}'.format(width))
        elif height and (height != 'auto'):
            thumbnail = thumbnails.get_thumbnail('h{0}'.format(height))
    if thumbnail is None:
        thumbnail = image
    if request is None:
        request = check_request()
    url = absolute_url(thumbnail, request)
    if timestamp:
        zdc = IZopeDublinCore(thumbnail, None)
        if zdc is None:
            timestamp = datetime.utcnow().timestamp()
        else:
            timestamp = zdc.modified.timestamp()    # pylint: disable=no-member
        url += '?_={0}'.format(timestamp)
    result = '<img src="{0}" class="{1}" alt="{2}" />'.format(url, img_class, alt)
    if css_class:
        result = '<div class="{0}">{1}</div>'.format(css_class, result)
    return result


def render_image(image, width=None, height=None, request=None, css_class='', img_class='',
                 timestamp=False, alt=''):
    # pylint: disable=too-many-arguments
    """Render image"""
    if ISVGImageFile.providedBy(image):
        return render_svg(image, width, height, request, css_class, img_class, alt)
    return render_img(image, width, height, request, css_class, img_class, timestamp, alt)



#
# Copyright (c) 2008-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.generations.evolve2 module

This module is doing a database scan of all registered blobs to add a reference to them
into blobs manager.
"""

import logging

from zope.intid import IIntIds

from pyams_file.interfaces import IBlobReferenceManager, IFile
from pyams_utils.registry import get_local_registry, get_utility, set_local_registry

__docformat__ = 'restructuredtext'


LOGGER = logging.getLogger('PyAMS (file)')


def evolve(site):
    """Evolve 2: create reference for all files blobs"""
    registry = get_local_registry()
    try:
        files = set()
        set_local_registry(site.getSiteManager())
        LOGGER.warning("Creating references to all blobs...")
        intids = get_utility(IIntIds)
        references = get_utility(IBlobReferenceManager)
        for ref in list(intids.refs.keys()):
            obj = intids.queryObject(ref)
            if IFile.providedBy(obj):
                blob = getattr(obj, '_blob', None)
                if blob is not None:
                    references.add_reference(blob, obj)
                LOGGER.debug(">>> updated blob reference for file {!r}".format(obj))
        LOGGER.warning("{} files updated".format(len(files)))
    finally:
        set_local_registry(registry)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.schema module

This module provides all custom file-related schema fields.
"""

from zope.interface import Attribute, Interface, implementer
from zope.schema import Bytes
from zope.schema.interfaces import IBytes, RequiredMissing, WrongType

from pyams_file.interfaces import IAudioFile, IBaseImageFile, IFile, IMediaFile, IVideoFile
from pyams_i18n.schema import I18nField, II18nField
from pyams_utils.interfaces.form import IDataManager, NOT_CHANGED, TO_BE_DELETED
from pyams_utils.registry import get_current_registry


__docformat__ = 'restructuredtext'


#
# Schema fields interfaces
#

class IThumbnailField(Interface):
    """Generic field interface with thumbnail"""


class IFileField(IBytes):
    """File object field interface"""

    schema = Attribute("Required value schema")


class IMediaField(IFileField):
    """Media file object field interface"""


class IThumbnailMediaField(IMediaField, IThumbnailField):
    """Media object field with thumbnail interface"""


class IImageField(IMediaField):
    """Image file object field interface"""


class IThumbnailImageField(IImageField, IThumbnailField):
    """Image object field with thumbnail interface"""


class IVideoField(IMediaField):
    """Video file field interface"""


class IThumbnailVideoField(IVideoField, IThumbnailField):
    """Video object field with thumbnail interface"""


class IAudioField(IMediaField):
    """Audio file field interface"""


#
# Custom schema fields
#

@implementer(IFileField)
class FileField(Bytes):
    """Custom field used to handle file-like properties"""

    schema = IFile

    def _validate(self, value):
        if value is TO_BE_DELETED:
            if self.required and not self.default:
                raise RequiredMissing
        elif value is NOT_CHANGED:
            return
        elif isinstance(value, tuple):
            try:
                filename, stream = value
                if not isinstance(filename, str):
                    raise WrongType(filename, str, '{0}.filename'.format(self.__name__))
                if not hasattr(stream, 'read'):
                    raise WrongType(stream, '<file-like object>', self.__name__)
            except ValueError:
                raise WrongType(value, tuple, self.__name__)
        elif not self.schema.providedBy(value):
            raise WrongType(value, self.schema, self.__name__)


@implementer(IMediaField)
class MediaField(FileField):
    """Custom field used to store media-like properties"""

    schema = IMediaFile


@implementer(IThumbnailMediaField)
class ThumbnailMediaField(MediaField):
    """Custom field used to store media properties with thumbnail selection"""


@implementer(IImageField)
class ImageField(MediaField):
    """Custom field used to handle image properties"""

    schema = IBaseImageFile


@implementer(IThumbnailImageField)
class ThumbnailImageField(ImageField):
    """Custom field used to handle images properties with square selection"""


@implementer(IVideoField)
class VideoField(MediaField):
    """Custom field used to store video file"""

    schema = IVideoFile


@implementer(IThumbnailVideoField)
class ThumbnailVideoField(VideoField):
    """Custom field used to store video properties with thumbnail selection"""


@implementer(IAudioField)
class AudioField(MediaField):
    """Custom field used to store audio file"""

    schema = IAudioFile


#
# I18n file fields
#

class II18nFileField(II18nField):
    """I18n file field marker interface"""


@implementer(II18nFileField)
class I18nFileField(I18nField):
    """I18n file field"""

    def __init__(self, key_type=None, value_type=None,
                 value_min_length=None, value_max_length=None, **kwargs):
        super(I18nFileField, self).__init__(
            value_type=FileField(min_length=value_min_length,
                                 max_length=value_max_length,
                                 required=False),
            **kwargs)

    def _validate(self, value):  # pylint: disable=too-many-branches
        if self.required:
            if self.default:
                return
            if not value:
                raise RequiredMissing
            has_value = False
            registry = get_current_registry()
            for lang, lang_value in value.items():
                # check for NOT_CHANGED value
                if lang_value is NOT_CHANGED:  # check for empty file value
                    adapter = registry.queryMultiAdapter((self.context, self), IDataManager)
                    if adapter is not None:
                        try:
                            old_value = adapter.query() or {}
                        except TypeError:  # can't adapt field context => new content?
                            old_value = None
                        else:
                            old_value = old_value.get(lang)
                    else:  # default data manager
                        old_value = getattr(self.context, self.__name__, {}).get(lang)
                    has_value = has_value or bool(old_value)
                    if has_value:
                        break
                elif lang_value is not TO_BE_DELETED:
                    has_value = True
                    break
            if not has_value:
                raise RequiredMissing
        for lang_value in value.values():
            if lang_value in (NOT_CHANGED, TO_BE_DELETED):
                return
            if isinstance(lang_value, tuple):
                try:
                    filename, stream = lang_value
                    if not isinstance(filename, str):
                        raise WrongType(filename, str, '{0}.filename'.format(self.__name__))
                    if not hasattr(stream, 'read'):
                        raise WrongType(stream, '<file-like object>', self.__name__)
                except ValueError:
                    raise WrongType(lang_value, tuple, self.__name__)
            elif not self.value_type.schema.providedBy(lang_value):
                raise WrongType(lang_value, self.value_type.schema, self.__name__)


class II18nMediaField(II18nFileField):
    """I18n media field marker interface"""


@implementer(II18nMediaField)
class I18nMediaField(I18nFileField):
    """I18n media field"""

    def __init__(self, key_type=None, value_type=None,
                 value_min_length=None, value_max_length=None, **kwargs):
        # pylint: disable=bad-super-call
        super(I18nFileField, self).__init__(
            value_type=MediaField(min_length=value_min_length,
                                  max_length=value_max_length,
                                  required=False),
            **kwargs)


class II18nThumbnailMediaField(II18nMediaField):
    """I18n field for media with thumbnail"""


@implementer(II18nThumbnailMediaField)
class I18nThumbnailMediaField(I18nMediaField):
    """I18n media field"""

    def __init__(self, key_type=None, value_type=None,
                 value_min_length=None, value_max_length=None, **kwargs):
        # pylint: disable=bad-super-call
        super(I18nFileField, self).__init__(
            value_type=ThumbnailMediaField(min_length=value_min_length,
                                           max_length=value_max_length,
                                           required=False),
            **kwargs)


class II18nImageField(II18nFileField):
    """I18n image field marker interface"""


@implementer(II18nImageField)
class I18nImageField(I18nMediaField):
    """I18n image field"""

    def __init__(self, key_type=None, value_type=None,
                 value_min_length=None, value_max_length=None, **kwargs):
        # pylint: disable=bad-super-call
        super(I18nFileField, self).__init__(
            value_type=ImageField(min_length=value_min_length,
                                  max_length=value_max_length,
                                  required=False),
            **kwargs)


class II18nThumbnailImageField(II18nImageField):
    """I18n field for image with thumbnail marker interface"""


@implementer(II18nThumbnailImageField)
class I18nThumbnailImageField(I18nImageField):
    """I18n field for image with thumbnail"""

    def __init__(self, key_type=None, value_type=None,
                 value_min_length=None, value_max_length=None, **kwargs):
        # pylint: disable=bad-super-call
        super(I18nFileField, self).__init__(
            value_type=ThumbnailImageField(min_length=value_min_length,
                                           max_length=value_max_length,
                                           required=False),
            **kwargs)


class II18nVideoField(II18nMediaField):
    """I18n video field marker interface"""


@implementer(II18nVideoField)
class I18nVideoField(I18nMediaField):
    """I18n video field"""

    def __init__(self, key_type=None, value_type=None,
                 value_min_length=None, value_max_length=None, **kwargs):
        # pylint: disable=bad-super-call
        super(I18nFileField, self).__init__(
            value_type=VideoField(min_length=value_min_length,
                                  max_length=value_max_length,
                                  required=False),
            **kwargs)


class II18nThumbnailVideoField(II18nVideoField):
    """I18n field for video with thumbnail marker interface"""


@implementer(II18nThumbnailVideoField)
class I18nThumbnailVideoField(I18nFileField):
    """I18n field for video with thumbnail"""

    def __init__(self, key_type=None, value_type=None,
                 value_min_length=None, value_max_length=None, **kwargs):
        # pylint: disable=bad-super-call
        super(I18nFileField, self).__init__(
            value_type=ThumbnailVideoField(min_length=value_min_length,
                                           max_length=value_max_length,
                                           required=False),
            **kwargs)


class II18nAudioField(II18nMediaField):
    """I18n audio field marker interface"""


@implementer(II18nAudioField)
class I18nAudioField(I18nMediaField):
    """I18n audio field"""

    def __init__(self, key_type=None, value_type=None,
                 value_min_length=None, value_max_length=None, **kwargs):
        # pylint: disable=bad-super-call
        super(I18nFileField, self).__init__(
            value_type=AudioField(min_length=value_min_length,
                                  max_length=value_max_length,
                                  required=False),
            **kwargs)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.archive.bz2 module

Bzip2 extraction module.
"""

import bz2

from pyams_file.archive import ArchiveExtractorBase
from pyams_file.archive.tar import TarArchiveExtractor
from pyams_file.file import get_magic_content_type
from pyams_file.interfaces.archive import IArchiveExtractor
from pyams_utils.registry import utility_config


__docformat__ = 'restructuredtext'


CHUNK_SIZE = 4096


@utility_config(name='application/x-bzip2', provides=IArchiveExtractor)
class BZip2ArchiveExtractor(ArchiveExtractorBase):
    """BZip2 file format archive extractor"""

    def _initialize(self, data, mode='r'):
        """Initialize extractor"""
        data = super(BZip2ArchiveExtractor, self)._initialize(data, mode)
        return data, bz2.BZ2Decompressor()

    def get_contents(self, data, mode='r'):
        """Extract archive contents"""
        data, extractor = self._initialize(data, mode)
        decompressed = extractor.decompress(data.read(CHUNK_SIZE))
        while not decompressed:
            decompressed = extractor.decompress(data.read(CHUNK_SIZE))
        mime_type = get_magic_content_type(decompressed[:CHUNK_SIZE])
        if mime_type == 'application/x-tar':
            tar = TarArchiveExtractor()
            data.seek(0)
            yield from tar.get_contents(data, 'r:bz2')
        else:
            while extractor.needs_input:
                decompressed += extractor.decompress(data.read(CHUNK_SIZE))
            yield decompressed, ''



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.interfaces.archive module

This module provides a single helper interface to handle archives.
"""

from zope.interface import Interface


__docformat__ = 'restructuredtext'


class IArchiveExtractor(Interface):
    """Archive contents extractor"""

    def get_contents(self, data, mode='r'):
        """Get iterator over archive contents

        Each iteration is a tuple containing data and file name.
        """



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.archive.zip module

ZIP files extraction module.
"""

import zipfile

from pyams_file.archive import ArchiveExtractorBase
from pyams_file.file import get_magic_content_type
from pyams_file.interfaces.archive import IArchiveExtractor
from pyams_utils.registry import query_utility, utility_config


__docformat__ = 'restructuredtext'


CHUNK_SIZE = 4096


@utility_config(name='application/zip', provides=IArchiveExtractor)
class ZipArchiveExtractor(ArchiveExtractorBase):
    """ZIP file format archive extractor"""

    zip_data = None

    def _initialize(self, data, mode='r'):
        """Initialize extractor"""
        data = super(ZipArchiveExtractor, self)._initialize(data, mode)
        return data, zipfile.ZipFile(data, mode=mode)

    def get_contents(self, data, mode='r'):
        """Extract archive contents"""
        data, extractor = self._initialize(data, mode)
        members = extractor.infolist()
        for member in members:
            filename = member.filename
            content = extractor.read(filename)
            if not content:  # skip empty files and directories
                continue
            mime_type = get_magic_content_type(content[:CHUNK_SIZE])
            archiver = query_utility(IArchiveExtractor, name=mime_type)
            if archiver is not None:
                yield from archiver.get_contents(content)
            else:
                yield content, filename



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""
Generic test cases for pyams_file doctests
"""

import os
import sys
from fnmatch import fnmatch

from persistent import Persistent
from zope.container.contained import Contained
from zope.interface import Interface, implementer

from pyams_file.property import FileProperty, I18nFileProperty
from pyams_file.schema import FileField, I18nAudioField, I18nFileField, I18nImageField, \
    I18nMediaField, \
    I18nThumbnailImageField, I18nThumbnailMediaField, I18nThumbnailVideoField, I18nVideoField, \
    ImageField


__docformat__ = 'restructuredtext'


def get_package_dir(value):
    """Get package directory"""
    package_dir = os.path.split(value)[0]
    if package_dir not in sys.path:
        sys.path.append(package_dir)
    return package_dir


def find_files(pattern, path):
    """Find files in given path matching name pattern"""
    for root, dirs, files in os.walk(path):  # pylint: disable=unused-variable
        for name in files:
            if fnmatch(name, pattern):
                yield (root, name)


#
# Tests classes
#

class IMyInterface(Interface):
    """Custom test interface"""
    data = FileField(title='File content', required=False)
    img_data = ImageField(title='Image content', required=False)
    required_data = FileField(title='Required field', required=True)



@implementer(IMyInterface)
class MyContent(Persistent, Contained):
    """Custom content class"""
    data = FileProperty(IMyInterface['data'])
    img_data = FileProperty(IMyInterface['img_data'])
    required_data = FileProperty(IMyInterface['required_data'])


class IMyI18nInterface(Interface):
    """Custom test interface with I18n field"""
    data = I18nFileField(title='File content', required=False)
    required_data = I18nFileField(title='Alternate content', required=True)
    media = I18nMediaField(title='Media field', required=False)
    thmedia = I18nThumbnailMediaField(title='Thumbnail media', required=False)
    image = I18nImageField(title='Image', required=False)
    thimage = I18nThumbnailImageField(title='Thumbnail image', required=False)
    video = I18nVideoField(title='Video', required=False)
    thvideo = I18nThumbnailVideoField(title='Thumbnail video', required=False)
    audio = I18nAudioField(title='Audio', required=False)


@implementer(IMyI18nInterface)
class MyI18nContent(Persistent, Contained):
    """Custom content class with I18n property"""
    data = I18nFileProperty(IMyI18nInterface['data'])
    required_data = I18nFileProperty(IMyI18nInterface['required_data'])



#
# Copyright (c) 2008-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.generations.evolve1 module

This generation module is looking for existing thumbnails and remove them from the database
to avois a bug in previous release which was leading to blob files never being correctly deleted
on database packing.
"""

import logging

from zope.intid import IIntIds

from pyams_file.interfaces import IMediaFile
from pyams_file.interfaces.thumbnail import IThumbnails
from pyams_utils.registry import get_local_registry, get_utility, set_local_registry

__docformat__ = 'restructuredtext'


LOGGER = logging.getLogger('PyAMS (file)')


def evolve(site):
    """Evolve 1: remove all images thumbnails to free blobs"""
    registry = get_local_registry()
    try:
        medias = set()
        set_local_registry(site.getSiteManager())
        LOGGER.warning("Removing all thumbnails from database to free unused blobs...")
        intids = get_utility(IIntIds)
        for ref in list(intids.refs.keys()):
            obj = intids.queryObject(ref)
            if IMediaFile.providedBy(obj):
                LOGGER.debug(">>> removing thumbnails for image {!r}".format(obj))
                thumbnails = IThumbnails(obj, None)
                if thumbnails is not None:
                    medias.add(obj)
                    thumbnails.clear_thumbnails()
        LOGGER.warning("Thumbnails cleanup is finished. Launch *zeopack* (for ZEO storage) "
                       "or *zodbpack* (for Relstorage) command to remove all unused blobs.")
        LOGGER.warning("{} images updated".format(len(medias)))
    finally:
        set_local_registry(registry)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.thumbnail module

This module is used to generate images thumbnails.
"""

import logging
import re

from BTrees import OOBTree  # pylint: disable=no-name-in-module
from persistent.dict import PersistentDict
from pyramid.events import subscriber
from pyramid.threadlocal import get_current_registry
from zope.interface import alsoProvides
from zope.lifecycleevent import IObjectRemovedEvent, ObjectAddedEvent, ObjectCreatedEvent, \
    ObjectRemovedEvent
from zope.location import locate
from zope.traversing.interfaces import ITraversable

from pyams_file.file import FileFactory
from pyams_file.interfaces import IFileModifiedEvent, IImageFile, IMediaFile
from pyams_file.interfaces.thumbnail import IThumbnailFile, IThumbnailer, IThumbnails, \
    IWatermarker
from pyams_utils.adapter import ContextAdapter, adapter_config, get_annotation_adapter
from pyams_utils.registry import query_utility
from pyams_utils.request import check_request


__docformat__ = 'restructuredtext'


LOGGER = logging.getLogger('PyAMS (file)')


THUMBNAIL_ANNOTATIONS_KEY = 'pyams_file.image.thumbnails'
THUMBNAIL_GEOMETRY_KEY = 'pyams_file.image.geometry'

THUMB_WIDTH = re.compile(r'^(?:\w+:)?w([0-9]+)$')
THUMB_HEIGHT = re.compile(r'^(?:\w+:)?h([0-9]+)$')
THUMB_SIZE = re.compile(r'^(?:\w+:)?([0-9]+)x([0-9]+)$')


@adapter_config(context=IImageFile, provides=IThumbnails)
class ImageThumbnailAdapter:
    """Image thumbnails adapter"""

    def __init__(self, image):
        self.image = image
        self.thumbnails = get_annotation_adapter(image, THUMBNAIL_ANNOTATIONS_KEY, OOBTree.OOBTree,
                                                 notify=False, locate=False)

    def get_image_size(self):
        """Get image size, in pixels"""
        return self.image.get_image_size()

    def get_thumbnail_size(self, thumbnail_name, forced=False):
        """Get size in pixels of requested thumbnail"""
        w_ratio, h_ratio = 1., 1.
        width, height = self.get_image_size()
        match = THUMB_WIDTH.match(thumbnail_name)
        if match:
            w = int(match.groups()[0])  # pylint: disable=invalid-name
            w_ratio = 1. * width / w
            h_ratio = 0.
        else:
            match = THUMB_HEIGHT.match(thumbnail_name)
            if match:
                h = int(match.groups()[0])  # pylint: disable=invalid-name
                w_ratio = 0.
                h_ratio = 1. * height / h
            else:
                match = THUMB_SIZE.match(thumbnail_name)
                if match:
                    groups = match.groups()
                    w = int(groups[0])  # pylint: disable=invalid-name
                    h = int(groups[1])  # pylint: disable=invalid-name
                    w_ratio = 1. * width / w
                    h_ratio = 1. * height / h
        if match:
            ratio = max(w_ratio, h_ratio)
            if not forced:
                ratio = max(ratio, 1.)
            return int(width / ratio), int(height / ratio)
        return None

    def get_geometry(self, selection_name):
        """Get geometry matching given selection name"""
        geometries = get_annotation_adapter(self.image, THUMBNAIL_GEOMETRY_KEY, default={})
        # get default geometry for custom thumbnails
        if ':' in selection_name:
            selection_name, options = selection_name.split(':', 1)
        else:
            options = None
        if selection_name in geometries:
            return geometries[selection_name]
        registry = check_request().registry
        thumbnailer = registry.queryAdapter(self.image, IThumbnailer, name=selection_name)
        if thumbnailer is not None:
            return thumbnailer.get_default_geometry(options)
        return None

    def set_geometry(self, selection_name, geometry):
        """Set geometry for given selection name"""
        geometries = get_annotation_adapter(self.image, THUMBNAIL_GEOMETRY_KEY, PersistentDict,
                                            notify=False, locate=False)
        geometries[selection_name] = geometry
        for current_thumbnail_name in list(self.thumbnails.keys()):
            if (current_thumbnail_name == selection_name) or \
                    current_thumbnail_name.startswith('{0}:'.format(selection_name)):
                self.delete_thumbnail(current_thumbnail_name)

    def clear_geometries(self):
        """Clear all stored geometries"""
        geometries = get_annotation_adapter(self.image, THUMBNAIL_GEOMETRY_KEY)
        if geometries is not None:
            for geometry_name in list(geometries.keys()):
                del geometries[geometry_name]

    def get_thumbnail_name(self, thumbnail_name, with_size=False):
        """Get name for given thumbnail name"""
        size = self.get_thumbnail_size(thumbnail_name)
        if size is not None:
            if with_size:
                return '{0}x{1}'.format(*size), size
            return '{0}x{1}'.format(*size)
        return None, None

    def get_selection(self, selection_name, format=None):
        # pylint: disable=redefined-builtin
        """Get thumbnail for given selection name"""
        LOGGER.debug(">>> Requested thumbnail selection: {}".format(selection_name))
        if selection_name in self.thumbnails:
            return self.thumbnails[selection_name]
        geometry = self.get_geometry(selection_name)
        if geometry == IThumbnailer(self.image).get_default_geometry():
            return self.image
        registry = get_current_registry()
        thumbnailer = registry.queryAdapter(self.image, IThumbnailer, name=selection_name)
        if thumbnailer is not None:
            selection = thumbnailer.create_thumbnail(geometry, format)
            if selection is not None:
                if isinstance(selection, tuple):
                    selection, format = selection
                else:
                    format = 'jpeg'
                selection = FileFactory(selection)
                alsoProvides(selection, IThumbnailFile)
                registry.notify(ObjectCreatedEvent(selection))
                self.thumbnails[selection_name] = selection
                selection_size = selection.get_image_size()
                locate(selection, self.image,
                       '++thumb++{0}:{1}x{2}.{3}'.format(selection_name,
                                                         selection_size[0],
                                                         selection_size[1],
                                                         format))
                LOGGER.debug("  > Generated thumbnail selection: {}".format(selection.__name__))
                registry.notify(ObjectAddedEvent(selection))
                return selection
        return None

    def get_thumbnail(self, thumbnail_name, format=None, watermark=None,
                      watermark_position='scale', watermark_opacity=1):
        # pylint: disable=redefined-builtin,too-many-arguments,too-many-locals,too-many-branches
        """Get thumbnail for given thumbnail name, which can provide selection and size"""
        LOGGER.debug(">>> Requested thumbnail: {}".format(thumbnail_name))
        # check for existing thumbnail
        if thumbnail_name in self.thumbnails:
            return self.thumbnails[thumbnail_name]
        # check for selection thumbnail, in "selection:size" format
        if ':' in thumbnail_name:
            selection_name, size = thumbnail_name.split(':', 1)
            selection = self.get_selection(selection_name)
            if selection is not None:
                thumbnails = IThumbnails(selection)
                return thumbnails.get_thumbnail(size)
        # check for matching one
        name, size = self.get_thumbnail_name(thumbnail_name, with_size=True)
        if name:
            if name in self.thumbnails:
                return self.thumbnails[name]
            # check for original image
            if size == self.get_image_size():
                return self.image
            # wee will look for default image thumbnailer
            thumbnailer_name = ''
            options = name
        else:
            if ':' in thumbnail_name:
                thumbnailer_name, options = thumbnail_name.split(':', 1)
            else:
                thumbnailer_name = thumbnail_name
                options = name = thumbnail_name
        # generate and store thumbnail
        registry = get_current_registry()
        thumbnailer = registry.queryAdapter(self.image, IThumbnailer, name=thumbnailer_name)
        if thumbnailer is not None:
            thumbnail_image = thumbnailer.create_thumbnail(options, format)
            if thumbnail_image is not None:
                if isinstance(thumbnail_image, tuple):
                    thumbnail_image, format = thumbnail_image
                else:
                    format = 'jpeg'
                # check watermark
                if watermark is not None:
                    watermarker = query_utility(IWatermarker)
                    if watermarker is not None:
                        thumbnail_image.seek(0)
                        thumbnail_image, format = watermarker.add_watermark(thumbnail_image,
                                                                            watermark,
                                                                            watermark_position,
                                                                            watermark_opacity)
                # create final image
                thumbnail_image = FileFactory(thumbnail_image)
                alsoProvides(thumbnail_image, IThumbnailFile)
                registry.notify(ObjectCreatedEvent(thumbnail_image))
                self.thumbnails[name] = thumbnail_image
                thumbnail_size = thumbnail_image.get_image_size()
                locate(thumbnail_image, self.image,
                       '++thumb++{0}{1}{2}x{3}.{4}'.format(thumbnailer_name,
                                                           ':' if thumbnailer_name else '',
                                                           thumbnail_size[0],
                                                           thumbnail_size[1],
                                                           format))
                LOGGER.debug("  < Generated thumbnail: {}".format(thumbnail_image.__name__))
                registry.notify(ObjectAddedEvent(thumbnail_image))
                return thumbnail_image
        return None

    def delete_thumbnail(self, thumbnail_name):
        """Delete given thumbnail"""
        if thumbnail_name in self.thumbnails:
            thumbnail_image = self.thumbnails[thumbnail_name]
            registry = get_current_registry()
            registry.notify(ObjectRemovedEvent(thumbnail_image))
            del self.thumbnails[thumbnail_name]
            LOGGER.debug(">>> Removed thumbnail: {}".format(thumbnail_name))

    def clear_thumbnails(self):
        """Clear all current image thumbnails"""
        # pylint: disable=expression-not-assigned
        [self.delete_thumbnail(thumbnail_name) for thumbnail_name in list(self.thumbnails.keys())]


@subscriber(IFileModifiedEvent, context_selector=IMediaFile)
@subscriber(IObjectRemovedEvent, context_selector=IMediaFile)
def handle_modified_image(event):
    """Clear thumbnails and selections when an image is updated or removed"""
    thumbnails = IThumbnails(event.object, None)
    if thumbnails is not None:
        thumbnails.clear_geometries()
        thumbnails.clear_thumbnails()


@adapter_config(name='thumb', context=IImageFile, provides=ITraversable)
class ThumbnailTraverser(ContextAdapter):
    """++thumb++ namespace traverser"""

    def traverse(self, name, furtherpath=None):  # pylint: disable=unused-argument
        """Traverse to given thumbnail name"""
        if '.' in name:
            thumbnail_name, format = name.rsplit('.', 1)  # pylint: disable=redefined-builtin
        else:
            thumbnail_name = name
            format = None  # pylint: disable=redefined-builtin
        # pylint: disable=assignment-from-no-return
        thumbnails = IThumbnails(self.context)
        if ':' in thumbnail_name:
            selection_name, thumbnail_name = thumbnail_name.split(':', 1)
            selection = thumbnails.get_selection(selection_name, format)
            if selection is not None:
                thumbnails = IThumbnails(selection)
        # pylint: disable=assignment-from-no-return
        return thumbnails.get_thumbnail(thumbnail_name, format)



#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.watermark module

This module is used to add watermarks to images.
"""

import os.path
from io import BytesIO, StringIO

from PIL import Image, ImageEnhance

from pyams_file.interfaces import IImageFile
from pyams_file.interfaces.thumbnail import IWatermarker
from pyams_utils.registry import utility_config


__docformat__ = 'restructuredtext'


@utility_config(provides=IWatermarker)
class ImageWatermarker:
    """Image watermarker utility"""

    @staticmethod
    def _reduce_opacity(image, opacity):
        """Returns an image with reduced opacity."""
        assert 0 <= opacity <= 1
        if image.mode != 'RGBA':
            image = image.convert('RGBA')
        else:
            image = image.copy()
        alpha = image.split()[3]
        alpha = ImageEnhance.Brightness(alpha).enhance(opacity)
        image.putalpha(alpha)
        return image

    def add_watermark(self, image, watermark, position='scale', opacity=1, format=None):
        # pylint: disable=redefined-builtin,too-many-arguments,too-many-branches
        """Adds a watermark to an image and return a new image"""
        # init image
        if IImageFile.providedBy(image):
            image = image.data
        if isinstance(image, bytes):
            image = BytesIO(image)
        elif isinstance(image, str) and not os.path.exists(image):
            image = StringIO(image)
        image = Image.open(image)
        # check format
        if not format:
            format = image.format
        format = format.upper()
        if format not in ('GIF', 'JPEG', 'PNG'):
            format = 'JPEG'
        # check RGBA mode
        if image.mode != 'RGBA':
            image = image.convert('RGBA')
        # init watermark
        if isinstance(watermark, str) and os.path.exists(watermark):
            watermark = Image.open(watermark)
        else:
            if IImageFile.providedBy(watermark):
                watermark = Image.open(StringIO(watermark.data))
            else:
                watermark = Image.open(watermark)
        if opacity < 1:
            watermark = self._reduce_opacity(watermark, opacity)
        # create a transparent layer the size of the image and draw the
        # watermark in that layer.
        layer = Image.new('RGBA', image.size, (0, 0, 0, 0))
        if position == 'tile':
            for y in range(0, image.size[1], watermark.size[1]):  # pylint: disable=invalid-name
                for x in range(0, image.size[0], watermark.size[0]):  # pylint: disable=invalid-name
                    layer.paste(watermark, (x, y))
        elif position == 'scale':
            # scale, but preserve the aspect ratio
            ratio = min(float(image.size[0]) / watermark.size[0],
                        float(image.size[1]) / watermark.size[1])
            w = int(watermark.size[0] * ratio)  # pylint: disable=invalid-name
            h = int(watermark.size[1] * ratio)  # pylint: disable=invalid-name
            watermark = watermark.resize((w, h))
            layer.paste(watermark, (int((image.size[0] - w) / 2), int((image.size[1] - h) / 2)))
        else:
            layer.paste(watermark, position)
        # composite the watermark with the layer
        new = BytesIO()
        composite = Image.composite(layer, image, layer)
        if format == 'JPEG':
            composite = composite.convert('RGB')
        composite.save(new, format)
        return new.getvalue(), format.lower()



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_file.interfaces module

This module provides all module interfaces and constants.
"""

from zope.annotation import IAttributeAnnotatable
from zope.interface import Interface, implementer
from zope.lifecycleevent import IObjectModifiedEvent, ObjectModifiedEvent
from zope.schema import Bytes, BytesLine, Choice, Text, TextLine


__docformat__ = 'restructuredtext'

from pyams_file import _


#
# Blobs references manager
#

class IBlobReferenceManager(Interface):
    """Blobs references manager

    This utility interface is used to manage references to blobs: each file contains a
    link to a ZODB "Blob" object which is used to store it's data; when a content is duplicated,
    all it's blobs references are updated but the blob itself is not duplicated to reduce space
    usage. As long as a blob data is not modified, the same content can be shared between several
    versions of a same content. So it's only when all references to a given blob have been removed
    that the blob file is deleted and can be garbaged collected.
    """

    def add_reference(self, blob, reference):
        """Add a reference to given blob"""

    def drop_reference(self, blob, reference):
        """Remove reference from given blob

        Blob is deleted if no more referenced.
        """


#
# Main file objects interfaces
#

class IFileModifiedEvent(IObjectModifiedEvent):
    """Modified file event interface"""


@implementer(IFileModifiedEvent)
class FileModifiedEvent(ObjectModifiedEvent):
    """Modified file event"""


class IFile(IAttributeAnnotatable):
    """File object interface"""

    content_type = BytesLine(title="Content type",
                             description="The content type identifies the type of content data",
                             required=False,
                             default=b'',
                             missing_value=b'')

    data = Bytes(title="Content data",
                 description="Actual file content",
                 required=False,
                 default=b'',
                 missing_value=b'')

    def get_size(self):
        """Returns the byte-size of object's data"""

    def get_blob(self, mode='r'):
        """Get Blob file associated with this object"""

    def add_blob_reference(self, reference):
        """Add a reference to file internal blob"""

    def free_blob(self):
        """Free blob associated with this object"""


class IMediaFile(IFile):
    """Multimedia file"""


class IBaseImageFile(IMediaFile):
    """Base image interface"""


class IImageFile(IBaseImageFile):
    """Image object interface"""

    def get_image_size(self):
        """Returns an (x, y) tuple describing image dimensions"""

    def resize(self, width, height):
        """Resize image to given dimensions"""

    def crop(self, x1, y1, x2, y2):  # pylint: disable=invalid-name
        """Crop image to given coordinates"""

    def rotate(self, angle=-90):
        """Rotate image, default to right"""


class ISVGImageFile(IBaseImageFile):
    """SVG file interface"""


class IResponsiveImage(Interface):
    """Responsive image marker interface"""


class IVideoFile(IMediaFile):
    """Video file interface"""


class IAudioFile(IMediaFile):
    """Audio file interface"""


class IFileInfo(Interface):
    """File extended information"""

    title = TextLine(title=_("Title"),
                     required=False)

    description = Text(title=_("Description"),
                       required=False)

    filename = TextLine(title=_("Save file as..."),
                        description=_("Name under which the file will be saved"),
                        required=False)

    language = Choice(title=_("Language"),
                      description=_("File's content language"),
                      vocabulary="PyAMS base languages",
                      required=False)


class IFileFieldContainer(IAttributeAnnotatable):
    """Marker interface for contents holding file properties"""


FILE_CONTAINER_ATTRIBUTES = 'pyams_file.file.attributes'
